#!/usr/bin/env python3
"""
Automatic summarissed report to condense results of an experiment.
It should NOT contain text beyond labels and titles.
The most efficient way to communicate information is (inteligently) colour-coded figures and tables.

The structure page-wise is as follows:
Page:Contents
1: Technical summary
2-4: Summarised signal results
5-End:
  Part 1: Extended results
  Part 2: Extended technicalities
"""

from agora.base import ParametersABC, ProcessABC


class Reporter(ProcessABC):
    """ """

    def __init__(self, parameters: ParametersABC):
        super().__init__(parameters)


class ReporterParameters(ParametersABC):
    """ """

    def __init__(self, technical: dict, results: dict, grouper_kws: dict):
        super().__init__()



#!/usr/bin/env python3

import pandas as pd
import umap
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


class umapembeddingParameters(ParametersABC):
    """
    Parameters for the 'umapembedding' process.
    """

    _defaults = {
        "n_neighbors": None,
        "min_dist": None,
        "n_components": None,
    }


class umapembedding(PostProcessABC):
    """
    Process to get UMAP embeddings from data/features.

    Methods
    -------
    run(signal: pd.DataFrame)
        Get UMAP embeddings.
    """

    def __init__(self, parameters: umapembeddingParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        """Get UMAP embeddings.

        This function is effectively a wrapper for umap.UMAP.

        Parameters
        ----------
        signal : pd.DataFrame
            Feature matrix.
        """
        mapper = umap.UMAP(
            n_neighbors=self.n_neighbors,
            min_dist=self.min_dist,
            n_components=self.n_components,
        ).fit(signal)
        return mapper.embedding_



import numpy as np
import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


class LineageProcessParameters(ParametersABC):
    """
    Parameters
    """

    _defaults = {}


class LineageProcess(PostProcessABC):
    """
    Lineage process that must be passed a (N,3) lineage matrix (where the coliumns are trap, mother, daughter respectively)
    """

    def __init__(self, parameters: LineageProcessParameters):
        super().__init__(parameters)

    def run(
        self,
    ):
        pass

    def filter_signal_cells(self, signal: pd.DataFrame):
        """
        Use casting to filter cell ids in signal and lineage
        """

        sig_ind = np.array(list(signal.index)).T[:, None, :]
        mo_av = (
            (self.lineage[:, :2].T[:, :, None] == sig_ind)
            .all(axis=0)
            .any(axis=1)
        )
        da_av = (
            (self.lineage[:, [0, 2]].T[:, :, None] == sig_ind)
            .all(axis=0)
            .any(axis=1)
        )

        return self.lineage[mo_av & da_av]



import numpy as np
import pandas as pd
from agora.abc import ParametersABC
from scipy import signal

from postprocessor.core.abc import PostProcessABC


class butterParameters(ParametersABC):
    """Parameters for the 'butter' process.

    Parameters for the 'butter' process.

    Attributes
    ----------
    order : int
        The order of the filter.
    critical_freqs : array_like
        The critical frequency or frequencies.  For lowpass and highpass
        filters, Wn is a scalar; for bandpass and bandstop filters, Wn is a
        length-2 sequence.  For a Butterworth filter, this is the point at which
        the gain drops to 1/sqrt(2) that of the passband (the “-3 dB point”).
        For digital filters, if fs is not specified, Wn units are normalized
        from 0 to 1, where 1 is the Nyquist frequency (Wn is thus in
        half cycles / sample and defined as 2*critical frequencies / fs).
        If fs is specified, Wn is in the same units as fs.  For analog filters,
        Wn is an angular frequency (e.g. rad/s).
    filter_type : {‘lowpass’, ‘highpass’, ‘bandpass’, ‘bandstop’}
        The type of filter. Default is ‘lowpass’.
    sampling_freq : float
        The sampling frequency of the digital system.
    """

    _defaults = {
        "order": 2,
        "critical_freqs": 1 / 350,
        "filter_type": "highpass",
        "sampling_freq": 1 / 5,
    }


class butter(PostProcessABC):
    """Process to apply Butterworth filter
    based on scipy.signal.butter

    Methods
    -------
    run(signal: pd.DataFrame)
        Apply Butterworth filter constructed according to user parameters
        to each time series in a DataFrame
    """

    def __init__(self, parameters: butterParameters):
        super().__init__(parameters)

    def butterfilter(self, timeseries):
        """Apply Butterworth filter to one time series"""
        # second-order-sections output
        # by default, using a digital filter
        sos = signal.butter(
            N=self.order,
            Wn=self.critical_freqs,
            btype=self.filter_type,
            fs=self.sampling_freq,
            output="sos",
        )
        # subtract time series by mean
        # otherwise the first couple time series will look like the acf,
        # which is not what we want
        # filter time series
        timeseries_norm = timeseries - np.mean(timeseries)
        return signal.sosfiltfilt(sos, timeseries_norm)

    def run(self, signal_df: pd.DataFrame):
        """Apply Butterworth filter

        Parameters
        ----------
        signal : pd.DataFrame
            Time series, with rows indicating individual time series (e.g. from
            each cell), and columns indicating time points.

        Returns
        -------
        signal_filtered : pd.DataFrame
            Filtered time series.

        """
        signal_filtered = signal_df.apply(
            self.butterfilter, axis=1, result_type="expand"
        )
        signal_filtered.columns = signal_df.columns
        return signal_filtered



import numpy as np
import pandas as pd
from agora.abc import ParametersABC
from sklearn import svm
from sklearn.decomposition import PCA
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import GridSearchCV, train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from postprocessor.core.abc import PostProcessABC


class miParameters(ParametersABC):
    """Parameters for the 'mi' process

    Parameters for the 'mi' process.

    Attributes
    ----------
    overtime: boolean (default: True)

        If True, calculate the mutual information as a function of the duration of
        the time series, by finding the mutuation information for all possible
        sub-time series that start from t= 0.


    n_bootstraps: int, optional (default: 100)

        The number of bootstraps used to estimate errors.


    ci: 1x2 array or list, optional (default: [0.25, 0.75])

        The lower and upper confidence intervals.

        E.g. [0.25, 0.75] for the interquartile range


    Crange: array, optional

        An array of potential values for the C parameter of the support vector machine
        and from which the optimal value of C will be chosen.

        If None, np.logspace(-3, 3, 10) is used. This range should be increased if
        the optimal C is one of the boundary values.

        See https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html


    gammarange: array, optional

        An array of potential values for the gamma parameter for the radial basis
        function kernel of the support vector machine and from which the optimal
        value of gamma will be chosen.

        If None, np.logspace(-3, 3, 10) is used. This range should be increased if
        the optimal gamma is one of the boundary values.

        See https://scikit-learn.org/stable/auto_examples/svm/plot_rbf_parameters.html


    train_test_split_seeding: boolean, optional (default: False)

        If True, force a random state for the train-test split in each bootstrap. This is
        useful in case the user requires reproducibility e.g. code testing.

        See https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html
    """

    _defaults = {
        "overtime": True,
        "n_bootstraps": 100,
        "ci": [0.25, 0.75],
        "Crange": None,
        "gammarange": None,
        "train_test_split_seeding": False,
    }


class mi(PostProcessABC):
    """
    Process to estimate power spectral density (autoregressive model).

    Methods
    -------
    run(signal: pd.DataFrame)
        Estimates the mutual information between classes of time series.
    """

    def __init__(
        self,
        parameters: miParameters,
    ):
        super().__init__(parameters)

    def run(self, signals):
        """
        Estimates the mutual information between classes of time series.

        Uses sklean to optimise a pipeline for classifying the individual time series,
        choosing the number of PCA components (3-7), the classifier - a support vector
        machine with either a linear or a radial basis function kernel - and its C and
        gamma parameters.

        Errors are found using bootstrapped datasets.

        Parameters:  signals: list of pandas.DataFrames

                        A list of DataFrames.  Each DataFrame stores a set of time series, with rows
                        indicating individual time series (e.g. from each cell), and columns
                        indicating time points.

        Returns:    res: array

                        Summary statistics from the bootstrapped datasets -- the median mutual
                        information and the 10% and 90% confidence limits.

                        If overtime is True, each row corresponds to a different duration of the time
                        series with the shortest duration, just the first time point, in the first row
                        and the longest duration, the entire time series, in the last row.
        """

        # default values
        if not np.any(self.Crange):
            self.Crange = np.logspace(-3, 3, 10)
        if not np.any(self.gammarange):
            self.gammarange = np.logspace(-3, 3, 10)

        # data is a list with one array of time series for different class
        data = [signal.to_numpy() for signal in signals]
        n_classes = len(data)
        Xo = np.vstack([timeseries for timeseries in data])
        y = np.hstack(
            [
                i * np.ones(timeseries.shape[0])
                for i, timeseries in enumerate(data)
            ]
        )

        if self.overtime:
            # loop over time series iteratively
            durations = np.arange(1, Xo.shape[1] + 1)
        else:
            # full time series only
            durations = [Xo.shape[1]]
        # array for results
        res = np.zeros((len(durations), 3))

        for j, duration in enumerate(durations):
            # slice of of time series
            # if verbose:
            #    print("duration of time series is", duration)
            X = Xo[:, :duration]

            # initialise sself.cikit-learn routines
            nPCArange = (
                range(1, X.shape[1] + 1) if X.shape[1] < 7 else [3, 4, 5, 6, 7]
            )
            params = [
                {"project__n_components": nPCArange},
                {
                    "classifier__kernel": ["linear"],
                    "classifier__C": self.Crange,
                },
                {
                    "classifier__kernel": ["rbf"],
                    "classifier__C": self.Crange,
                    "classifier__gamma": self.gammarange,
                },
            ]
            pipe = Pipeline(
                [
                    ("project", PCA()),
                    ("rescale", StandardScaler()),
                    ("classifier", svm.SVC()),
                ]
            )

            # find best params for pipeline
            grid_pipeline = GridSearchCV(pipe, params, n_jobs=-1, cv=5)
            grid_pipeline.fit(X, y)
            # if verbose:
            #    print(grid_pipeline.best_estimator_)
            pipe.set_params(**grid_pipeline.best_params_)

            # find mutual information for each bootstrapped dataset
            mi = np.empty(self.n_bootstraps)
            for i in range(self.n_bootstraps):
                # force random state, useful for code testing/reproducibility
                if self.train_test_split_seeding:
                    random_state = i
                else:
                    random_state = None
                X_train, X_test, y_train, y_test = train_test_split(
                    X,
                    y,
                    test_size=0.25,
                    random_state=random_state,
                )
                # run classifier use optimal params
                pipe.fit(X_train, y_train)
                y_predict = pipe.predict(X_test)
                # estimate mutual information
                p_y = 1 / n_classes
                p_yhat_given_y = confusion_matrix(
                    y_test, y_predict, normalize="true"
                )
                p_yhat = np.sum(p_y * p_yhat_given_y, 0)
                h_yhat = -np.sum(
                    p_yhat[p_yhat > 0] * np.log2(p_yhat[p_yhat > 0])
                )
                log2_p_yhat_given_y = np.ma.log2(p_yhat_given_y).filled(0)
                h_yhat_given_y = -np.sum(
                    p_y * np.sum(p_yhat_given_y * log2_p_yhat_given_y, 1)
                )
                mi[i] = h_yhat - h_yhat_given_y

            # summary statistics - median and confidence intervals
            res[j, :] = [
                np.median(mi),
                np.sort(mi)[int(np.min(self.ci) * self.n_bootstraps)],
                np.sort(mi)[int(np.max(self.ci) * self.n_bootstraps)],
            ]
            # if verbose:
            #    print(f"median MI= {res[j,0]:.2f} [{res[j,1]:.2f}, {res[j,2]:.2f}]")
        return res



import numpy as np
import pandas as pd
from agora.abc import ParametersABC
from scipy.signal import savgol_filter

from postprocessor.core.abc import PostProcessABC


class savgolParameters(ParametersABC):
    """
    Parameters

        window : int (odd)
            Window length of datapoints. Must be odd and smaller than x
        polynom : int
            The order of polynom used. Must be smaller than the window size
    """

    _defaults = {"window": 7, "polynom": 3}


class savgol(PostProcessABC):
    """
    Apply Savitzky-Golay filter (works with NaNs, but it might return
    NaN regions).
    """

    def __init__(self, parameters: savgolParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        try:
            post_savgol = pd.DataFrame(
                savgol_filter(
                    signal, self.parameters.window, self.parameters.polynom
                ),
                index=signal.index,
                columns=signal.columns,
            )
        except Exception as e:
            print(e)

            def savgol_on_srs(x):
                return self.non_uniform_savgol(
                    x.index,
                    x.values,
                    self.parameters.window,
                    self.parameters.polynom,
                )

            post_savgol = signal.apply(savgol_on_srs, 1).apply(pd.Series)
        return post_savgol


def non_uniform_savgol(x, y, window: int, polynom: int):
    """
    Applies a Savitzky-Golay filter to y with non-uniform spacing
    as defined in x

    This is based on https://dsp.stackexchange.com/questions/1676/savitzky-golay-smoothing-filter-for-not-equally-spaced-data
    The borders are interpolated like scipy.signal.savgol_filter would do

    source: https://dsp.stackexchange.com/a/64313

    Parameters
    ----------
    x : array_like
        List of floats representing the x values of the data
    y : array_like
        List of floats representing the y values. Must have same length
        as x
    window : int (odd)
        Window length of datapoints. Must be odd and smaller than x
    polynom : int
        The order of polynom used. Must be smaller than the window size

    Returns
    -------
    np.array of float
        The smoothed y values
    """
    _raiseif(
        len(x) != len(y),
        '"x" and "y" must be of the same size',
        ValueError,
    )
    _raiseif(
        len(x) < window,
        "The data size must be larger than the window size",
        ValueError,
    )
    _raiseif(
        not isinstance(window, int),
        '"window" must be an integer',
        TypeError,
    )
    _raiseif(window % 2, 'The "window" must be an odd integer', ValueError)

    _raiseif(
        not isinstance(polynom, int),
        '"polynom" must be an integer',
        TypeError,
    )

    _raiseif(
        polynom >= window,
        '"polynom" must be less than "window"',
        ValueError,
    )

    half_window = window // 2
    polynom += 1

    # Initialize variables
    A = np.empty((window, polynom))  # Matrix
    tA = np.empty((polynom, window))  # Transposed matrix
    t = np.empty(window)  # Local x variables
    y_smoothed = np.full(len(y), np.nan)

    # Start smoothing
    for i in range(half_window, len(x) - half_window, 1):
        # Center a window of x values on x[i]
        for j in range(0, window, 1):
            t[j] = x[i + j - half_window] - x[i]

        # Create the initial matrix A and its transposed form tA
        for j in range(0, window, 1):
            r = 1.0
            for k in range(0, polynom, 1):
                A[j, k] = r
                tA[k, j] = r
                r *= t[j]

        # Multiply the two matrices
        tAA = np.matmul(tA, A)

        # Invert the product of the matrices
        tAA = np.linalg.inv(tAA)

        # Calculate the pseudoinverse of the design matrix
        coeffs = np.matmul(tAA, tA)

        # Calculate c0 which is also the y value for y[i]
        y_smoothed[i] = 0
        for j in range(0, window, 1):
            y_smoothed[i] += coeffs[0, j] * y[i + j - half_window]

        # If at the end or beginning, store all coefficients for the polynom
        if i == half_window:
            first_coeffs = np.zeros(polynom)
            for j in range(0, window, 1):
                for k in range(polynom):
                    first_coeffs[k] += coeffs[k, j] * y[j]
        elif i == len(x) - half_window - 1:
            last_coeffs = np.zeros(polynom)
            for j in range(0, window, 1):
                for k in range(polynom):
                    last_coeffs[k] += coeffs[k, j] * y[len(y) - window + j]

    # Interpolate the result at the left border
    for i in range(0, half_window, 1):
        y_smoothed[i] = 0
        x_i = 1
        for j in range(0, polynom, 1):
            y_smoothed[i] += first_coeffs[j] * x_i
            x_i *= x[i] - x[half_window]

    # Interpolate the result at the right border
    for i in range(len(x) - half_window, len(x), 1):
        y_smoothed[i] = 0
        x_i = 1
        for j in range(0, polynom, 1):
            y_smoothed[i] += last_coeffs[j] * x_i
            x_i *= x[i] - x[-half_window - 1]

    return y_smoothed


def _raiseif(cond, msg="", exc=AssertionError):
    if cond:
        raise exc(msg)



#!/usr/bin/env python3

import numpy as np
import pandas as pd
from agora.abc import ParametersABC
from scipy.signal import find_peaks

from postprocessor.core.abc import PostProcessABC


class findpeaksParameters(ParametersABC):
    """
    Parameters for the 'findpeaks' process.

    Attributes
    ----------
    distance: number
        Required minimal horizontal distance (>= 1) in samples between
        neighbouring peaks. Smaller peaks are removed first until the condition
        is fulfilled for all remaining peaks.

    prominence: number or ndarray or sequence
        Required prominence of peaks. Either a number, None, an array matching x
        or a 2-element sequence of the former. The first element is always
        interpreted as the minimal and the second, if supplied, as the maximal
        required prominence.
    """

    _defaults = {
        "height": None,
        "threshold": None,
        "distance": 10,
        "prominence": 0.035,
        "width": None,
        "wlen": None,
        "rel_height": 0.5,
        "plateau_size": None,
    }


class findpeaks(PostProcessABC):
    """
    Process to find peaks inside a signal.

    Methods
    -------
    _find_peaks_mask(timeseries: sequence, distance: number, prominence: number
        or ndarray or sequence)
        Find peaks of a time series and returns a binary mask locating these peaks
    run(signal: pd.DataFrame)
        Find peaks in a dataframe of time series.
    """

    def __init__(self, parameters: findpeaksParameters):
        super().__init__(parameters)

    def _find_peaks_mask(
        self,
        timeseries,
        height,
        threshold,
        distance,
        prominence,
        width,
        wlen,
        rel_height,
        plateau_size,
    ):
        """Find peaks of a time series and returns a binary mask locating these peaks

        Parameters
        ----------
        timeseries : sequence
            Time series with peaks.
        distance : number
            Required minimal horizontal distance in samples between neighbouring
            peaks.
        prominence : number or ndarray or sequence
            Required prominence of peaks.
        """
        peak_indices = find_peaks(
            timeseries,
            height=height,
            threshold=threshold,
            distance=distance,
            prominence=prominence,
            width=width,
            wlen=wlen,
            rel_height=rel_height,
            plateau_size=plateau_size,
        )[0]
        mask = np.zeros(len(timeseries), dtype=int)
        mask[peak_indices] = 1
        return mask

    def run(self, signal: pd.DataFrame):
        """Find peaks in a dataframe of time series.

        Find peaks of a dataframe of time series. This function is effectively a
        wrapper for scipy.signal.find_peaks.

        Parameters
        ----------
        signal : pd.DataFrame
            Time series, with rows indicating individual time series (e.g. from
            each cell), and columns indicating time points.
        """
        mask_df = signal.apply(
            lambda x: self._find_peaks_mask(
                x,
                height=self.height,
                threshold=self.threshold,
                distance=self.distance,
                prominence=self.prominence,
                width=self.width,
                wlen=self.wlen,
                rel_height=self.rel_height,
                plateau_size=self.plateau_size,
            ),
            axis=1,
            result_type="expand",
        )
        mask_df.columns = signal.columns
        return mask_df



from itertools import product

import igraph as ig
import leidenalg
import numpy as np
import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


class leidenParameters(ParametersABC):
    """
    Parameters
    """

    _defaults = {}


class leiden(PostProcessABC):
    """
    leiden algorithm applied to a dataframe with features.
    """

    def __init__(self, parameters: leidenParameters):
        super().__init__(parameters)

    def run(self, features: pd.DataFrame):
        # Generate euclidean distance matrix
        distances = np.linalg.norm(
            features.values - features.values[:, None], axis=2
        )
        ind = [
            "_".join([str(y) for y in x[1:]])
            for x in features.index.to_flat_index()
        ]
        source, target = zip(*product(ind, ind))
        df = pd.DataFrame(
            {
                "source": source,
                "target": target,
                "distance": distances.flatten(),
            }
        )
        df = df.loc[df["source"] != df["target"]]
        g = ig.Graph.DataFrame(df, directed=False)

        return leidenalg.find_partition(g, leidenalg.ModularityVertexPartition)



#!/usr/bin/env python3
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages


def dummyplot():
    plt.plot(np.random.randint(10, size=5), np.random.randint(10, size=5))


plt1 = dummyplot()
plt2 = dummyplot()
pp = PdfPages("foo.pdf")
for i in [plt1, plt2]:
    pp.savefig(i)
pp.close()



import numpy as np
import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


class detrendParameters(ParametersABC):
    """Parameters for the 'detrend' process.

    Parameters for the 'detrend' process.

    Attributes
    ----------
    window : int
        Size of sliding window.
    """

    _defaults = {"window": 16}


class detrend(PostProcessABC):
    """Process to detrend using sliding window

    Methods
    -------
    run(signal: pd.DataFrame)
        Detrend each time series in a dataframe using a specified sliding window
    """

    def __init__(self, parameters: detrendParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        """Detrend using sliding window

        Detrend each time series in a dataframe using a specified sliding window

        Parameters
        ----------
        signal : pd.DataFrame
            Time series, with rows indicating individual time series (e.g. from
            each cell), and columns indicating time points.

        Returns
        -------
        signal_norm : pd.DataFrame
            Detrended time series.

        """
        # Compute moving average
        signal_movavg = signal.rolling(
            window=self.window, center=True, axis=1
        ).mean()
        # Detrend: subtract normalised time series by moving average
        signal_detrend = signal.subtract(signal_movavg)
        return signal_detrend.dropna(axis=1)  # Remove columns with NaNs



#!/usr/bin/env python3

import re
from abc import ABC, abstractproperty
from collections import Counter
from pathlib import Path, PosixPath
from typing import Dict, List, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from agora.io.signal import Signal
from pathos.multiprocessing import Pool


class Grouper(ABC):
    """Base grouper class."""

    files = []

    def __init__(self, dir: Union[str, PosixPath]):
        path = Path(dir)
        self.name = path.name
        assert path.exists(), "Dir does not exist"
        self.files = list(path.glob("*.h5"))
        assert len(self.files), "No valid h5 files in dir"
        self.load_signals()

    def load_signals(self):
        self.signals = {f.name[:-3]: Signal(f) for f in self.files}

    @property
    def fsignal(self) -> list:
        return list(self.signals.values())[0]

    @property
    def ntimepoints(self) -> int:
        return max([s.ntimepoints for s in self.signals.values()])

    @property
    def tintervals(self) -> float:
        tintervals = set([s.tinterval[0] / 60 for s in self.signals.values()])
        assert (
            len(tintervals) == 1
        ), "Not all signals have the same time interval"

        return max(tintervals)

    @property
    def siglist(self) -> None:
        return self.fsignal.siglist

    @property
    def siglist_grouped(self) -> None:
        if not hasattr(self, "_siglist_grouped"):
            self._siglist_grouped = Counter(
                [x for s in self.signals.values() for x in s.siglist]
            )

        for s, n in self._siglist_grouped.items():
            print(f"{s} - {n}")

    @property
    def datasets(self) -> None:
        """Print available datasets in first Signal instance."""
        return self.fsignal.datasets

    @abstractproperty
    def group_names(self):
        pass

    def concat_signal(
        self,
        path: str,
        reduce_cols: bool = None,
        axis: int = 0,
        pool: int = 0,
        mode="retained",
        **kwargs,
    ):
        """Concatenate a single signal.

        Parameters
        ----------
        path : str
            signal address within h5py file.
        reduce_cols : bool
            Whether or not to collapse columns into a single one.
        axis : int
            Concatenation axis.
        pool : int
            Number of threads used. If 0 or None only one core is used.
        **kwargs : key, value pairings
            Named arguments to pass to concat_ind_function.

        Examples
        --------
        FIXME: Add docs.
        """
        if not path.startswith("/"):
            path = "/" + path

        group_names = self.group_names

        # Check the path is in a given signal
        sitems = {k: v for k, v in self.signals.items() if path in v.siglist}
        nsignals_dif = len(self.signals) - len(sitems)
        if nsignals_dif:
            print(
                f"Grouper:Warning: {nsignals_dif} signals do not contain"
                f" channel {path}"
            )

        if pool or pool is None:
            if pool is None:
                pool = 8
            with Pool(pool) as p:
                signals = p.map(
                    lambda x: concat_signal_ind(
                        path, group_names, x[0], x[1], mode=mode, **kwargs
                    ),
                    sitems.items(),
                )
        else:
            signals = []
            for name, signal in sitems.items():
                print(name)
                signals.append(
                    concat_signal_ind(
                        path, group_names, name, signal, **kwargs
                    )
                )

        errors = [k for s, k in zip(signals, self.signals.keys()) if s is None]
        signals = [s for s in signals if s is not None]
        if len(errors):
            print("Warning: Positions contain errors {errors}")
            assert len(signals), "All datasets contain errors"
        sorted = pd.concat(signals, axis=axis).sort_index()
        if reduce_cols:
            sorted = sorted.apply(np.nanmean, axis=1)
            spath = path.split("/")
            sorted.name = "_".join([spath[1], spath[-1]])

        return sorted

    @property
    def ntraps(self):
        for pos, s in self.signals.items():
            with h5py.File(s.filename, "r") as f:
                print(pos, f["/trap_info/trap_locations"].shape[0])

    def traplocs(self):
        d = {}
        for pos, s in self.signals.items():
            with h5py.File(s.filename, "r") as f:
                d[pos] = f["/trap_info/trap_locations"][()]
        return d


class MetaGrouper(Grouper):
    """Group positions using metadata's 'group' number."""

    pass


class NameGrouper(Grouper):
    """Group a set of positions using a subsection of the name."""

    def __init__(self, dir, by=None):
        super().__init__(dir=dir)

        if by is None:
            by = (0, -4)
        self.by = by

    @property
    def group_names(self):
        if not hasattr(self, "_group_names"):
            self._group_names = {}
            for name in self.signals.keys():
                self._group_names[name] = name[self.by[0] : self.by[1]]

        return self._group_names

    def aggregate_multisignals(self, paths=None, **kwargs):
        aggregated = pd.concat(
            [
                self.concat_signal(path, reduce_cols=np.nanmean, **kwargs)
                for path in paths
            ],
            axis=1,
        )
        # ph = pd.Series(
        #     [
        #         self.ph_from_group(x[list(aggregated.index.names).index("group")])
        #         for x in aggregated.index
        #     ],
        #     index=aggregated.index,
        #     name="media_pH",
        # )
        # self.aggregated = pd.concat((aggregated, ph), axis=1)

        return aggregated


class phGrouper(NameGrouper):
    """Grouper for pH calibration experiments where all surveyed media pH
    values are within a single experiment."""

    def __init__(self, dir, by=(3, 7)):
        super().__init__(dir=dir, by=by)

    def get_ph(self) -> None:
        self.ph = {gn: self.ph_from_group(gn) for gn in self.group_names}

    @staticmethod
    def ph_from_group(group_name: int) -> float:
        if group_name.startswith("ph_"):
            group_name = group_name[3:]

        return float(group_name.replace("_", "."))

    def aggregate_multisignals(self, paths: list) -> pd.DataFrame:
        """Accumulate multiple signals."""

        aggregated = pd.concat(
            [
                self.concat_signal(path, reduce_cols=np.nanmean)
                for path in paths
            ],
            axis=1,
        )
        ph = pd.Series(
            [
                self.ph_from_group(
                    x[list(aggregated.index.names).index("group")]
                )
                for x in aggregated.index
            ],
            index=aggregated.index,
            name="media_pH",
        )
        aggregated = pd.concat((aggregated, ph), axis=1)

        return aggregated


def concat_signal_ind(
    path: str,
    group_names: List[str],
    group: str,
    signal: str,
    mode: str = "retained",
    **kwargs,
) -> pd.DataFrame:
    """Core function that handles retrieval of an individual signal, applies
    filtering if requested and adjusts indices."""
    if mode == "retained":
        combined = signal.retained(path, **kwargs)
    if mode == "mothers":
        raise (NotImplementedError)
    elif mode == "raw":
        combined = signal.get_raw(path, **kwargs)
    elif mode == "families":
        combined = signal[path]
    combined["position"] = group
    combined["group"] = group_names[group]
    combined.set_index(["group", "position"], inplace=True, append=True)
    combined.index = combined.index.swaplevel(-2, 0).swaplevel(-1, 1)

    return combined
    # except:
    #     return None


class MultiGrouper:
    """Wrap results from multiple experiments stored as folders inside a
    folder."""

    def __init__(self, source: Union[str, list]):
        if isinstance(source, str):
            source = Path(source)
            self.exp_dirs = list(source.glob("*"))
        else:
            self.exp_dirs = [Path(x) for x in source]
        self.groupers = [NameGrouper(d) for d in self.exp_dirs]
        for group in self.groupers:
            group.load_signals()

    @property
    def siglist(self) -> None:
        for gpr in self.groupers:
            print(gpr.siglist_grouped)

    @property
    def sigtable(self) -> pd.DataFrame:
        """Generate a matrix containing the number of datasets for each signal
        and experiment."""

        def regex_cleanup(x):
            x = re.sub(r"\/extraction\/", "", x)
            x = re.sub(r"\/postprocessing\/", "", x)
            x = re.sub(r"\/np_max", "", x)

            return x

        if not hasattr(self, "_sigtable"):
            raw_mat = [
                [s.siglist for s in gpr.signals.values()]
                for gpr in self.groupers
            ]
            siglist_grouped = [
                Counter([x for y in grp for x in y]) for grp in raw_mat
            ]

            nexps = len(siglist_grouped)
            sigs_idx = list(
                set([y for x in siglist_grouped for y in x.keys()])
            )
            sigs_idx = [regex_cleanup(x) for x in sigs_idx]

            nsigs = len(sigs_idx)

            sig_matrix = np.zeros((nsigs, nexps))
            for i, c in enumerate(siglist_grouped):
                for k, v in c.items():
                    sig_matrix[sigs_idx.index(regex_cleanup(k)), i] = v

            sig_matrix[sig_matrix == 0] = np.nan
            self._sigtable = pd.DataFrame(
                sig_matrix,
                index=sigs_idx,
                columns=[x.name for x in self.exp_dirs],
            )
        return self._sigtable

    def sigtable_plot(self) -> None:
        """Plot number of signals for all available experiments.

        Examples
        --------
        FIXME: Add docs.
        """
        ax = sns.heatmap(self.sigtable, cmap="viridis")
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=10,
            ha="right",
            rotation_mode="anchor",
        )
        plt.show()

    def aggregate_signal(
        self,
        signals: Union[str, list],
        **kwargs,
    ) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        """Aggregate signals from multiple Groupers (and thus experiments)

        Parameters
        ----------
        signals : Union[str, list]
            string or list of strings indicating the signal(s) to fetch.
        **kwargs : keyword arguments to pass to Grouper.concat_signal
            Customise the filters and format to fetch signals.

        Returns
        -------
        Union[pd.DataFrame, Dict[str, pd.DataFrame]]
            DataFrame or list of DataFrames

        Examples
        --------
        FIXME: Add docs.
        """
        if isinstance(signals, str):
            signals = [signals]

        sigs = {s: [] for s in signals}
        for s in signals:
            for grp in self.groupers:
                try:
                    sigset = grp.concat_signal(s, **kwargs)
                    new_idx = pd.MultiIndex.from_tuples(
                        [(grp.name, *x) for x in sigset.index],
                        names=("experiment", *sigset.index.names),
                    )

                    sigset.index = new_idx
                    sigs[s].append(sigset)
                except Exception as e:
                    print("Grouper {} failed: {}".format(grp.name, e))
                    # raise (e)

        concated = {
            name: pd.concat(multiexp_sig)
            for name, multiexp_sig in sigs.items()
        }
        if len(concated) == 1:
            concated = list(concated.values())[0]

        return concated



#!/usr/bin/env python3

import numpy as np
import pandas as pd
from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC


def df_extend_nan(df, width):
    """Extend a DataFrame to the left by a number of columns and fill with NaNs

    Assumes column names are sequential integers from 0
    """
    num_rows, _ = df.shape
    nan_df = pd.DataFrame(
        np.full([num_rows, width], np.nan),
        index=df.index,
    )
    out_df = pd.concat([nan_df, df], axis=1)
    _, out_num_cols = out_df.shape
    out_df.columns = list(range(out_num_cols))
    return out_df


def df_shift(df, list_index, shift_list):
    """Shifts each row of each DataFrame by a list of shift intervals

    Assumes all DataFrames have the same indices (and therefore the same number of rows)
    """
    for index, shift in zip(list_index, shift_list):
        df.loc[index, :] = df.loc[index, :].shift(periods=shift)
    return df


class alignParameters(ParametersABC):
    """
    Parameters for the 'align' process.

    Attributes
    ----------
    slice_before_first_event: bool
        Whether to discard the parts of signals that occur before the first
        event being aligned.  For example, whether to discard flavin
        fluorescence before the first birth event, after aligning by the first
        birth event.
    events_at_least: int
        Specifies the number of events required for each cell.  For example, if
        events_at_least is 2, then it will discard time series (from the DataFrame)
        that have less than 2 events.  As a more pratical example: discarding
        flavin time series that derive from cells with less than 2 buddings
        identified.
    """

    _defaults = {
        "slice_before_first_event": True,
        "events_at_least": 1,
    }


class align(PostProcessABC):
    """
    Process to align a signal by corresponding events.

    For example, aligning flavin fluorescence time series by the first birth
    event of the cell each time series is derived from.

    Methods
    -------
    run(trace_df: pd.DataFrame, mask_df: pd.DataFrame)
        Align signals by events.
    """

    def __init__(self, parameters: alignParameters):
        super().__init__(parameters)

    # Not sure if having two DataFrame inputs fits the paradigm, but having the
    # mask_df be a parameter is a bit odd as it doesn't set the behaviour of the
    # process.
    def run(self, trace_df: pd.DataFrame, mask_df: pd.DataFrame):
        """Align signals by events.

        Parameters
        ----------
        trace_df : pd.DataFrame
            Signal time series, with rows indicating individual time series
            (e.g. from each cell), and columns indicating time points.
        mask_df : pd.DataFrame
            Event time series/mask, with rows indicating individual cells and
            columns indicating time points. The values of each element are
            either 0 or 1 -- 0 indicating the absence of the event, and 1
            indicating the presence of the event. Effectively, this DataFrame is
            like a mask. For example, this DataFrame can indicate when birth
            events are identified for each cell in a dataset.
        """
        # Remove cells that have less than or equal to events_at_least events,
        # i.e. if events_at_least = 1, then cells that have no birth events are
        # deleted.
        event_mask = mask_df.apply(
            lambda x: np.sum(x) >= self.events_at_least, axis=1
        )
        mask_df = mask_df.iloc[event_mask.to_list()]

        # Match trace and event signals by index, e.g. cellID
        # and discard the cells they don't have in common
        common_index = trace_df.index.intersection(mask_df.index)
        trace_aligned = trace_df.loc[common_index]
        mask_aligned = mask_df.loc[common_index]

        # Identify first event and define shift
        shift_list = []
        for index in common_index:
            event_locs = np.where(mask_df.loc[index].to_numpy() == 1)[0]
            if event_locs.any():
                shift = event_locs[0]
            else:
                shift = 0
            shift_list.append(shift)
        shift_list = np.array(shift_list)

        # Shifting

        # Remove bits of traces before first event
        if self.slice_before_first_event:
            # minus sign in front of shift_list to shift to the left
            mask_aligned = df_shift(
                mask_aligned, common_index.to_list(), -shift_list
            )
            trace_aligned = df_shift(
                trace_aligned, common_index.to_list(), -shift_list
            )
        # Do not remove bits of traces before first event
        else:
            # Add columns to left, filled with NaNs
            max_shift = np.max(shift_list)
            mask_aligned = df_extend_nan(mask_aligned, max_shift)
            trace_aligned = df_extend_nan(trace_aligned, max_shift)
            # shift each
            mask_aligned = df_shift(
                mask_aligned, common_index.to_list(), -shift_list
            )
            trace_aligned = df_shift(
                trace_aligned, common_index.to_list(), -shift_list
            )

        return trace_aligned, mask_aligned



#!/usr/bin/env python3

from abc import ABC


class BasePlotter(ABC):
    """Base class for plotting handler classes"""

    def __init__(self, trace_name, unit_scaling, xlabel, plot_title):
        """Common attributes"""
        self.trace_name = trace_name
        self.unit_scaling = unit_scaling

        self.xlabel = xlabel
        self.ylabel = None
        self.plot_title = plot_title

    def plot(self, ax):
        """Template for drawing on provided Axes"""
        ax.set_ylabel(self.ylabel)
        ax.set_xlabel(self.xlabel)
        ax.set_title(self.plot_title)
        # Derived classes extends this with plotting functions


# TODO: something about the plotting functions at the end of the modules.
# Decorator?
